# -*- coding: utf-8 -*-
"""
Copyright 2016 William La Cava

license: GNU/GPLv3

"""
import numpy as np
import copy
import pdb
from sklearn.metrics import r2_score
from .population import stacks_2_eqns

def tournament(individuals,tourn_size, num_selections=None):
    """conducts tournament selection of size tourn_size"""
    winners = []
    locs = []
    if num_selections is None:
        num_selections = len(individuals)

    for i in np.arange(num_selections):
        # sample pool with replacement
        pool_i = np.random.choice(len(individuals),size=tourn_size)
        pool = []
        for i in pool_i:
            pool.append(np.mean(individuals[i].fitness))
        # winner
        locs.append(pool_i[np.argmin(pool)])
        winners.append(copy.deepcopy(individuals[locs[-1]]))

    return winners,locs

def lexicase(individuals, num_selections=None, epsilon = False, survival = False):
    """conducts lexicase selection for de-aggregated fitness vectors"""
    if num_selections is None:
        num_selections = len(individuals)
    winners = []
    locs = []

    if epsilon: # use epsilon lexicase selection
        # calculate epsilon thresholds based on median absolute deviation (MAD)
        mad_for_case = np.empty([len(individuals[0].fitness_vec),1])
        global_best_val_for_case = np.empty([len(individuals[0].fitness_vec),1])
        for i in np.arange(len(individuals[0].fitness_vec)):
            mad_for_case[i] = mad(np.asarray(list(map(lambda x: x.fitness_vec[i], individuals))))
            global_best_val_for_case[i] = min(map(lambda x: x.fitness_vec[i], individuals))
        # convert fitness values to pass/fail based on epsilon distance
        for I in individuals:
            fail_condition = np.array(I.fitness_vec > global_best_val_for_case[:,0] + mad_for_case[:,0]) #[f > global_best_val_for_case+mad_for_case for f in I.fitness_vec]
            I.fitness_vec = fail_condition.astype(int)

    for i in np.arange(num_selections):

        candidates = individuals
        can_locs = range(len(individuals))
        cases = list(np.arange(len(individuals[0].fitness_vec)))
        np.random.shuffle(cases)
        # pdb.set_trace()
        while len(cases) > 0 and len(candidates) > 1:
            # get best fitness for case among candidates
            # print("candidates:",stacks_2_eqns(candidates),"locations:",can_locs)
            # print("fitnesses for case "+str(cases[0])+":",[x.fitness_vec[cases[0]] for x in candidates])
            best_val_for_case = min([x.fitness_vec[cases[0]] for x in candidates])
            # print("best_val_for_case:",best_val_for_case)
            # filter individuals without an elite fitness on this case
            # tmp_c,tmp_l = zip(*((x,l) for x,l in zip(candidates,can_locs) if x.fitness_vec[cases[0]] == best_val_for_case))
            candidates,can_locs = zip(*((x,l) for x,l in zip(candidates,can_locs) if x.fitness_vec[cases[0]] == best_val_for_case))
            cases.pop(0)

        choice = np.random.randint(len(candidates))
        winners.append(copy.deepcopy(candidates[choice]))
        locs.append(can_locs[choice])
        if survival: # filter out winners from remaining selection pool
            individuals = list(filter(lambda x: x.stack != candidates[choice].stack, individuals))

    return winners, locs

def epsilon_lexicase(individuals, num_selections=None, survival = False):
    """conducts lexicase selection for de-aggregated fitness vectors"""
    if num_selections is None:
        num_selections = len(individuals)
    winners = []
    locs = []
    individual_locs = range(len(individuals))
    # calculate epsilon thresholds based on median absolute deviation (MAD)
    mad_for_case = np.empty([len(individuals[0].fitness_vec),1])
    # global_best_val_for_case = np.empty([len(individuals[0].fitness_vec),1])
    for i in np.arange(len(individuals[0].fitness_vec)):
        mad_for_case[i] = mad(np.asarray(list(map(lambda x: x.fitness_vec[i], individuals))))
        # global_best_val_for_case[i] = min(map(lambda x: x.fitness_vec[i], individuals))
    # convert fitness values to pass/fail based on epsilon distance
    # for I in individuals:
    #     fail_condition = np.array(I.fitness_vec > global_best_val_for_case[:,0] + mad_for_case[:,0]) #[f > global_best_val_for_case+mad_for_case for f in I.fitness_vec]
    #     I.fitness_bool = fail_condition.astype(int)

    for i in np.arange(num_selections):

        candidates = individuals
        can_locs = individual_locs
        cases = list(np.arange(len(individuals[0].fitness_vec)))
        np.random.shuffle(cases)
        # pdb.set_trace()
        while len(cases) > 0 and len(candidates) > 1:
            # get best fitness for case among candidates
            # print("candidates:",stacks_2_eqns(candidates),"locations:",can_locs)
            # print("fitnesses for case "+str(cases[0])+":",[x.fitness_vec[cases[0]] for x in candidates])
            best_val_for_case = min([x.fitness_vec[cases[0]] for x in candidates])
            # print("best_val_for_case:",best_val_for_case)
            # filter individuals without an elite fitness on this case
            # tmp_c,tmp_l = zip(*((x,l) for x,l in zip(candidates,can_locs) if x.fitness_vec[cases[0]] == best_val_for_case))
            candidates,can_locs = zip(*((x,l) for x,l in zip(candidates,can_locs) if x.fitness_vec[cases[0]] <= best_val_for_case + mad_for_case[cases[0]]))
            cases.pop(0)

        choice = np.random.randint(len(candidates))
        winners.append(copy.deepcopy(candidates[choice]))
        locs.append(can_locs[choice])
        if survival: # filter out winners from remaining selection pool
            # individuals = list(filter(lambda x: x.stack != candidates[choice].stack, individuals))
            try:
                individuals, individual_locs = zip(*[(x,l) for x,l in zip(individuals,individual_locs) if x.stack != candidates[choice].stack])
            except ValueError: # there are only clones left
                break

    while len(winners) < num_selections:
        winners.append(copy.deepcopy(individuals[0]))
        locs.append(individual_locs[0])

    return winners, locs

def mad(x, axis=None):
    """median absolute deviation statistic"""
    return np.median(np.abs(x - np.median(x, axis)), axis)

def deterministic_crowding(parents,offspring,X_parents,X_offspring):
    """deterministic crowding implementation (for non-steady state).
    offspring compete against the parent they are most similar to, here defined as
    the parent they are most correlated with.
    the offspring only replace their parent if they are more fit.
    """

    # try:

    # get children locations produced from crossover
    cross_children = [i for i,o in enumerate(offspring) if len(o.parentid) > 1]
    # order offspring so that they are lined up with their most similar parent
    for c1,c2 in zip(cross_children[::2], cross_children[1::2]):
        # get parent locations
        p_loc = [j for j,p in enumerate(parents) if p.id in offspring[c1].parentid]
        if len(p_loc) != 2:
            pdb.set_trace()
        # if child is more correlated with its non-root parent
        if r2_score(X_parents[p_loc[0]],X_offspring[c1]) + r2_score(X_parents[p_loc[1]],X_offspring[c2]) < r2_score(X_parents[p_loc[0]],X_offspring[c2]) + r2_score(X_parents[p_loc[1]],X_offspring[c1]):
            # swap offspring
            offspring[c1],offspring[c2] = offspring[c2],offspring[c1]

    survivors = []
    survivor_index = []

    for i,(p,o) in enumerate(zip(parents,offspring)):
        if p.fitness >= o.fitness:
            survivors.append(copy.deepcopy(p))
            survivor_index.append(i)
        else:
            survivors.append(copy.deepcopy(o))
            survivor_index.append(i+len(parents))

    # except:
    #     pdb.set_trace()
    # return survivors along with their indices
    return survivors, survivor_index



"""
Copyright 2016 William La Cava

This file is part of the FEW library.

The FEW library is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

The FEW library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
the FEW library. If not, see http://www.gnu.org/licenses/.

"""
from few.population import *
from few.selection import *
from few import FEW
import numpy as np
# unit tests for selection methods.
def test_tournament_shapes():
    """test_selection.py: tournament selection returns correct shape"""

    few = FEW(seed_with_ml=False,population_size=257)
    few.term_set = [node('x',loc=0)]
    pop = few.init_pop()
    offspring,locs = tournament(pop.individuals,2)
    assert len(offspring) == 257

    offspring,locs = tournament(pop.individuals,5)
    assert len(offspring) == 257

    # smaller popsize than tournament size
    few = FEW(seed_with_ml=False,population_size=2)
    few.term_set = [node('x',loc=0)]
    pop = few.init_pop()
    offspring,locs = tournament(pop.individuals,5)
    assert len(offspring) == 2;

def test_lexicase_shapes():
    """test_selection.py: lexicase selection returns correct shape"""
    few = FEW(seed_with_ml=False,population_size=257)
    few.term_set = [node('x',loc=0)]
    pop = few.init_pop()
    offspring,locs = lexicase(pop.individuals)
    assert len(offspring) == 257

    # smaller popsize than tournament size
    few = FEW(seed_with_ml=False,population_size=2)
    few.term_set = [node('x',loc=0)]
    pop = few.init_pop()
    offspring,locs = lexicase(pop.individuals)
    assert len(offspring) == 2;

def test_epsilon_lexicase_shapes():
    """test_selection.py: epsilon lexicase selection returns correct shape"""

    few = FEW(seed_with_ml=False,population_size=257)
    few.term_set = [node('x',loc=0)]
    pop = few.init_pop()
    offspring,locs = lexicase(pop.individuals, epsilon=True)
    assert len(offspring) == 257

    # smaller popsize than tournament size
    few = FEW(seed_with_ml=False,population_size=2)
    few.term_set = [node('x',loc=0)]
    pop = few.init_pop()
    offspring,locs = lexicase(pop.individuals,epsilon=True)
    assert len(offspring) == 2;
    assert len(locs) == 2;

def test_lexicase_survival_shapes():
    """test_selection.py: lexicase survival returns correct shape"""
    # func_set = [node('+'), node('-'), node('*'), node('/'), node('sin'),
    #                  node('cos'), node('exp'),node('log'), node('^2'),
    #                  node('^3'), node('sqrt')]
    # terminal set
    term_set = []
    n_features = 3
    # numbers represent column indices of features
    # for i in np.arange(n_features):
    #     term_set.append(node('x',loc=i)) # features
    term_set = [node('x',loc=i) for i in np.arange(n_features)]
        # term_set.append(('erc',0,np.random.rand())) # ephemeral random constants

    few = FEW(seed_with_ml=False,population_size=257)
    few.term_set = term_set
    pop = few.init_pop()

    for i in pop.individuals:
        i.fitness_vec = list(np.random.rand(10,1))

    offspring,locs = lexicase(pop.individuals,num_selections=100,survival=True)
    assert len(offspring) == 100

    # smaller popsize than tournament size
    ew = FEW(seed_with_ml=False,population_size=2)
    few.term_set = term_set
    pop = few.init_pop()
    for i in pop.individuals:
        i.fitness_vec = np.random.rand(10,1)
    offspring,locs = lexicase(pop.individuals,num_selections=1,survival=True)
    assert len(offspring) == 1;



# -*- coding: utf-8 -*-

"""
Copyright 2016 William La Cava

This file is part of the FEW library.

License: GPLv3
"""

from ._version import __version__
from .few import FEW, main
# from .few import evaluation
# from .few import selection
# from .few import variation



"""
Copyright 2016 William La Cava

This file is part of the FEW library.

The FEW library is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

The FEW library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
the FEW library. If not, see http://www.gnu.org/licenses/.

"""
# test FEW methods
from few.few import FEW
from sklearn.datasets import load_boston, load_iris
from sklearn.metrics import r2_score
from sklearn.linear_model import LassoLarsCV
from sklearn.model_selection import train_test_split
import pandas as pd
import numpy as np

def test_few_fit_shapes():
    """test_few.py: fit and predict return correct shapes """
    np.random.seed(202)
    # load example data
    boston = load_boston()
    d = pd.DataFrame(data=boston.data)
    print("feature shape:",boston.data.shape)

    learner = FEW(generations=1, population_size=5,
                mutation_rate=0.2, crossover_rate=0.8,
                ml = LassoLarsCV(), min_depth = 1, max_depth = 3,
                sel = 'epsilon_lexicase', tourn_size = 2, random_state=0, verbosity=0,
                disable_update_check=False, fit_choice = 'mse')

    score = learner.fit(boston.data[:300], boston.target[:300])
    print("learner:",learner._best_estimator)
    yhat_test = learner.predict(boston.data[300:])
    test_score = learner.score(boston.data[300:],boston.target[300:])
    print("train score:",score,"test score:",test_score,
    "test r2:",r2_score(boston.target[300:],yhat_test))
    assert yhat_test.shape == boston.target[300:].shape


def test_few_at_least_as_good_as_default():
    """test_few.py: few performs at least as well as the default ML """
    np.random.seed(1006987)
    boston = load_boston()
    d = np.column_stack((boston.data,boston.target))
    np.random.shuffle(d)
    features = d[:,0:-1]
    target = d[:,-1]

    print("feature shape:",boston.data.shape)

    learner = FEW(generations=1, population_size=5,
                mutation_rate=1, crossover_rate=1,
                ml = LassoLarsCV(), min_depth = 1, max_depth = 3,
                sel = 'tournament', fit_choice = 'r2',tourn_size = 2, random_state=0, verbosity=0,
                disable_update_check=False)

    learner.fit(features[:300], target[:300])
    few_score = learner.score(features[:300], target[:300])
    test_score = learner.score(features[300:],target[300:])

    lasso = LassoLarsCV()
    lasso.fit(learner._training_features,learner._training_labels)
    lasso_score = lasso.score(features[:300], target[:300])
    print("few score:",few_score,"lasso score:",lasso_score)
    print("few test score:",test_score,"lasso test score:",lasso.score(features[300:],target[300:]))
    assert few_score >= lasso_score

    print("lasso coefficients:",lasso.coef_)

    # assert False
def test_few_classification():
    """test_few.py: tests default classification settings"""
    np.random.seed(42)
    X, y = load_iris(return_X_y=True)
    train,test = train_test_split(np.arange(X.shape[0]), train_size=0.75, test_size=0.25)
    few = FEW(classification=True,population_size='1x',generations=10)
    few.fit(X[train],y[train])

    print('train score:', few.score(X[train],y[train]))
    print('test score:', few.score(X[test],y[test]))

    # test boolean output
    few = FEW(classification=True,otype='b',population_size='2x',seed_with_ml=False,
              generations=10)
    np.random.seed(42)
    few.fit(X[train],y[train])

    print('train score:', few.score(X[train],y[train]))
    print('test score:', few.score(X[test],y[test]))
    few.print_model()



"""
Copyright 2016 William La Cava

This file is part of the FEW library.

License: GPL 3
"""
from .test_variation import test_cross_makes_valid_program, test_mutate_makes_valid_program
from .test_population import test_pop_shape, test_pop_init
from .test_selection import *
from .test_evaluation import test_out_shapes, test_out_is_correct, test_calc_fitness_shape
from .test_few import test_few_fit_shapes, test_few_at_least_as_good_as_default, test_few_classification



# -*- coding: utf-8 -*-
"""
Copyright 2016 William La Cava

license: GNU/GPLv3

"""

import argparse
from ._version import __version__
from .evaluation import out, calc_fitness
from .population import *
from .variation import cross, mutate
from .selection import *

from sklearn.base import BaseEstimator
from sklearn.linear_model import LassoLarsCV, LogisticRegression, SGDClassifier
from sklearn.svm import SVR, LinearSVR, SVC, LinearSVC
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.model_selection import train_test_split
from sklearn.metrics import r2_score, accuracy_score
from sklearn.preprocessing import Imputer
from DistanceClassifier import DistanceClassifier
import numpy as np
import pandas as pd
import warnings
import copy
import itertools as it
import pdb
# from update_checker import update_check
# from joblib import Parallel, delayed
from tqdm import tqdm
import uuid
# import multiprocessing as mp
# NUM_THREADS = mp.cpu_count()



class FEW(BaseEstimator):
    """FEW uses GP to find a set of transformations from the original feature space
    that produces the best performance for a given machine learner.
    """
    update_checked = False

    def __init__(self, population_size=50, generations=100,
                 mutation_rate=0.5, crossover_rate=0.5,
                 ml = None, min_depth = 1, max_depth = 2, max_depth_init = 2,
                 sel = 'epsilon_lexicase', tourn_size = 2, fit_choice = None, op_weight = False,
                 seed_with_ml = True, erc = False, random_state=np.random.randint(9999999), verbosity=0, scoring_function=None,
                 disable_update_check=False,elitism=True, boolean = False,classification=False,clean=False,
                 track_diversity=False,mdr=False,otype='f'):
                # sets up GP.

        # Save params to be recalled later by get_params()
        self.params = locals()  # Must be placed before any local variable definitions
        self.params.pop('self')

        # # Do not prompt the user to update during this session if they ever disabled the update check
        # if disable_update_check:
        #     FEW.update_checked = True
        #
        # # Prompt the user if their version is out of date
        # if not disable_update_check and not FEW.update_checked:
        #     update_check('FEW', __version__)
        #     FEW.update_checked = True

        self._best_estimator = None
        self._training_features = None
        self._training_labels = None
        self._best_inds = None

        self.population_size = population_size
        self.generations = generations
        self.mutation_rate = mutation_rate
        self.crossover_rate = crossover_rate
        self.min_depth = min_depth
        self.max_depth = max_depth
        self.max_depth_init = max_depth_init
        self.sel = sel
        self.tourn_size = tourn_size
        self.fit_choice = fit_choice
        self.op_weight = op_weight
        self.seed_with_ml = seed_with_ml
        self.erc = erc
        self.random_state = random_state
        self.verbosity = verbosity
        self.scoring_function = scoring_function
        self.gp_generation = 0
        self.elitism = elitism
        self.max_fit = 99999999.666
        self.boolean = boolean
        self.classification = classification
        self.clean = clean
        self.ml = ml
        self.track_diversity = track_diversity
        self.mdr = mdr
        self.otype = otype

        # if otype is b, boolean functions must be turned on
        if self.otype=='b':
            self.boolean = True
        # else:
        #     self.otype = 'f'


        # pdb.set_trace()
        # instantiate sklearn estimator according to specified machine learner
        if not self.ml:
            if self.classification:
                self.ml = LogisticRegression(solver='sag')
            else:
                self.ml = LassoLarsCV()
        if not self.scoring_function:
            if self.classification:
                self.scoring_function = accuracy_score
            else:
                self.scoring_function = r2_score

        # set default fitness metrics for various learners
        if not self.fit_choice:
            self.fit_choice =  {
                            #regression
                            type(LassoLarsCV()): 'mse',
                            type(SVR()): 'mae',
                            type(LinearSVR()): 'mae',
                            type(KNeighborsRegressor()): 'mse',
                            type(DecisionTreeRegressor()): 'mse',
                            type(RandomForestRegressor()): 'mse',
                            #classification
                            type(SGDClassifier()): 'silhouette',
                            type(LogisticRegression()): 'silhouette',
                            type(SVC()): 'silhouette',
                            type(LinearSVC()): 'silhouette',
                            type(RandomForestClassifier()): 'silhouette',
                            type(DecisionTreeClassifier()): 'silhouette',
                            type(DistanceClassifier()): 'silhouette',
                            type(KNeighborsClassifier()): 'silhouette',
            }[type(self.ml)]


        # Columns to always ignore when in an operator
        self.non_feature_columns = ['label', 'group', 'guess']

        # function set
        self.func_set = [node('+'), node('-'), node('*'), node('/'), node('sin'),
                         node('cos'), node('exp'),node('log'), node('^2'),
                         node('^3'), node('sqrt')]


        # if boolean operators are included but the output type is set to float, then
        # # include the if and if-else operations that allow use of both stacks
        # if self.boolean and self.otype=='f':
        #     self.func_set += [
        #     {'name:','if','arity':2,'in_type':}
        #     ]
        # terminal set
        self.term_set = []
        # diversity
        self.diversity = []

    def fit(self, features, labels):
        """Fit model to data"""
        np.random.seed(self.random_state)
        # setup data
        # imputation
        if self.clean:
            features = self.impute_data(features)
        # Train-test split routine for internal validation
        ####
        train_val_data = pd.DataFrame(data=features)
        train_val_data['labels'] = labels
        # print("train val data:",train_val_data[::10])
        new_col_names = {}
        for column in train_val_data.columns.values:
            if type(column) != str:
                new_col_names[column] = str(column).zfill(10)
        train_val_data.rename(columns=new_col_names, inplace=True)
        # internal training/validation split
        train_i, val_i = train_test_split(train_val_data.index,
                                                             stratify=None,
                                                             train_size=0.75,
                                                             test_size=0.25)

        x_t = train_val_data.loc[train_i].drop('labels',axis=1).values
        x_v = train_val_data.loc[val_i].drop('labels',axis=1).values
        y_t = train_val_data.loc[train_i, 'labels'].values
        y_v = train_val_data.loc[val_i, 'labels'].values

        # Store the training features and classes for later use
        self._training_features = x_t
        self._training_labels = y_t
        ####

        # set population size
        if type(self.population_size) is str:
            if 'x' in self.population_size: #
                self.population_size = int(float(self.population_size[:-1])*features.shape[1])
            else:
                self.population_size = int(self.population_size)

        if self.verbosity >0: print("population size:",self.population_size)
        # print few settings
        if self.verbosity > 1:
            for arg in self.get_params():
                print('{}\t=\t{}'.format(arg, self.get_params()[arg]))
            print('')

        # initial model
        self._best_estimator = copy.deepcopy(self.ml.fit(x_t,y_t))
        self._best_score = self._best_estimator.score(x_v,y_v)
        if self.verbosity > 2: print("initial estimator size:",self._best_estimator.coef_.shape)
        if self.verbosity > 0: print("initial ML CV: {:1.3f}".format(self._best_score))

        # create terminal set
        for i in np.arange(x_t.shape[1]):
            # dictionary of node name, arity, feature column index, output type and input type
            self.term_set.append(node('x',loc=i)) # features
            # add ephemeral random constants if flag
            if self.erc:
                self.term_set.append(node('k',value=np.random.rand())) # ephemeral random constants

        # edit function set if boolean
        if self.boolean or self.otype=='b': # include boolean functions
            self.func_set += [node('!'), node('&'), node('|'), node('=='),
                            node('>_f'), node('<_f'), node('>=_f'), node('<=_f'),
                            node('>_b'), node('<_b'), node('>=_b'), node('<=_b'),
                            node('xor_b'), node('xor_f')]

        # add mdr if specified
        if self.mdr:
            self.func_set += [node('mdr2')]

        # use de-aggregated fitness metrics for lexicase selection
        if "lexicase" in self.sel and ("_vec" not in self.fit_choice or "_rel" not in self.fit_choice):
            self.fit_choice += "_vec"
        # Create initial population
        pop = self.init_pop(self._training_features.shape[0])
        # check that uuids are unique in population
        uuids = [p.id for p in pop.individuals]
        if len(uuids) != len(set(uuids)):
            pdb.set_trace()
        # Evaluate the entire population
        # X represents a matrix of the population outputs (number os samples x population size)
        # single thread
        pop.X = self.transform(x_t,pop.individuals,y_t)
        # parallel:
        # pop.X = np.asarray(Parallel(n_jobs=-1)(delayed(out)(I,x_t,self.otype,y_t) for I in pop.individuals), order = 'F')
        # pdb.set_trace()
        # calculate fitness of individuals
        # fitnesses = list(map(lambda I: fitness(I,y_t,self.ml),pop.X))
        fitnesses = calc_fitness(pop.X,y_t,self.fit_choice)
        # pdb.set_trace()
        # max_count = 0;
        # pdb.set_trace()
        # max_fit = self.max_fit
        # while len([np.mean(f) for f in fitnesses if np.mean(f) < max_fit and np.mean(f)>=0])<self.population_size and max_count < 100:
        #     pop = self.init_pop()
        #     pop.X = self.transform(x_t,pop.individuals,y_t)
        #     fitnesses = calc_fitness(pop.X,y_t,self.fit_choice)
        #
        #     max_count+= 1
        # print("fitnesses:",fitnesses)
        # Assign fitnesses to inidividuals in population
        for ind, fit in zip(pop.individuals, fitnesses):
            if isinstance(fit,(list,np.ndarray)): # calc_fitness returned raw fitness values
                fit[fit < 0] = self.max_fit
                fit[np.isnan(fit)] = self.max_fit
                fit[np.isinf(fit)] = self.max_fit
                ind.fitness_vec = fit
                ind.fitness = np.mean(ind.fitness_vec)
            else:
                ind.fitness = np.nanmin([fit,self.max_fit])
        # pdb.set_trace()

        #with Parallel(n_jobs=10) as parallel:
        ####################
        ### Main GP loop
        self.diversity=[]
        # progress bar
        pbar = tqdm(total=self.generations,disable = self.verbosity==0,desc='Internal CV: {:1.3f}'.format(self._best_score))
        # for each generation g
        for g in np.arange(self.generations):

            if self.track_diversity:
                self.get_diversity(pop.X)

            if self.verbosity > 1: print(".",end='')
            if self.verbosity > 1: print(str(g)+".)",end='')
            # if self.verbosity > 1: print("population:",stacks_2_eqns(pop.individuals))
            if self.verbosity > 2: print("pop fitnesses:", ["%0.2f" % x.fitness for x in pop.individuals])
            if self.verbosity > 1: print("median fitness pop: %0.2f" % np.median([x.fitness for x in pop.individuals]))
            if self.verbosity > 1: print("best fitness pop: %0.2f" % np.min([x.fitness for x in pop.individuals]))
            if self.verbosity > 1 and self.track_diversity: print("feature diversity: %0.2f" % self.diversity[-1])
            if self.verbosity > 1: print("ml fitting...")
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                try:
                    if len(self.valid_loc(pop.individuals)) > 0:
                        self.ml.fit(pop.X[self.valid_loc(pop.individuals),:].transpose(),y_t)
                    else:
                        self.ml.fit(pop.X.transpose(),y_t)

                except ValueError as detail:
                    print("warning: ValueError in ml fit. X.shape:",pop.X[self.valid_loc(pop.individuals),:].transpose().shape,"y_t shape:",y_t.shape)
                    print("First ten entries X:",pop.X[self.valid_loc(pop.individuals),:].transpose()[:10])
                    print("First ten entries y_t:",y_t[:10])
                    print("equations:",stacks_2_eqns(pop.individuals))
                    print("FEW parameters:",self.get_params())
                    if self.verbosity > 1: print("---\ndetailed error message:",detail)
                    raise(ValueError)

            # if self.verbosity > 1: print("number of non-zero regressors:",self.ml.coef_.shape[0])
            # keep best model
            try:
                if len(self.valid_loc(pop.individuals)) > 0:
                    tmp = self.ml.score(self.transform(x_v,pop.individuals)[self.valid_loc(pop.individuals),:].transpose(),y_v)
                else:
                    tmp = self.ml.score(self.transform(x_v,pop.individuals).transpose(),y_v)
            except Exception:
                tmp = 0

            if self.verbosity > 1: print("current ml validation score:",tmp)


            if tmp > self._best_score:
                self._best_estimator = copy.deepcopy(self.ml)
                self._best_score = tmp
                self._best_inds = pop.individuals[:]
                if self.verbosity > 1: print("updated best internal validation score:",self._best_score)

            offspring = []

            if self.verbosity > 2: print("copying new pop...")
            # clone individuals for offspring creation
            if self.sel == 'lasso':
                # for lasso, filter individuals with 0 coefficients
                offspring = copy.deepcopy(list(x for i,x in zip(self.ml.coef_, self.valid(pop.individuals)) if  i != 0))
            else:
                offspring = copy.deepcopy(self.valid(pop.individuals))

            if self.elitism: # keep a copy of the elite individual
                elite_index = np.argmin([x.fitness for x in pop.individuals])
                elite = copy.deepcopy(pop.individuals[elite_index])

            # Apply crossover and mutation on the offspring
            if self.verbosity > 2: print("variation...")
            for child1, child2 in it.zip_longest(offspring[::2], offspring[1::2],fillvalue=None):

                if np.random.rand() < self.crossover_rate and child2 != None:
                # crossover
                    cross(child1.stack, child2.stack, self.max_depth)
                    # update ids
                    child1.parentid = [child1.id,child2.id]
                    child1.id = uuid.uuid4()
                    child2.parentid = [child1.id,child2.id]
                    child2.id = uuid.uuid4()
                    # set default fitness
                    child1.fitness = -1
                    child2.fitness = -1
                elif child2 == None:
                # single mutation
                    mutate(child1.stack,self.func_set,self.term_set)
                    # update ids
                    child1.parentid = [child1.id]
                    child1.id = uuid.uuid4()
                    # set default fitness
                    child1.fitness = -1
                else:
                #double mutation
                    mutate(child1.stack,self.func_set,self.term_set)
                    mutate(child2.stack,self.func_set,self.term_set)
                    # update ids
                    child1.parentid = [child1.id]
                    child1.id = uuid.uuid4()
                    child2.parentid = [child2.id]
                    child2.id = uuid.uuid4()
                    # set default fitness
                    child1.fitness = -1
                    child2.fitness = -1

            while len(offspring) < self.population_size:
                #make new offspring to replace the invalid ones
                offspring.append(Ind())
                make_program(offspring[-1].stack,self.func_set,self.term_set,np.random.randint(self.min_depth,self.max_depth+1),self.otype)
                offspring[-1].stack = list(reversed(offspring[-1].stack))

            # evaluate offspring
            if self.verbosity > 2: print("output...")
            X_offspring = self.transform(x_t,offspring)
            #parallel:
            # X_offspring = np.asarray(Parallel(n_jobs=-1)(delayed(out)(O,x_t,y_t,self.otype) for O in offspring), order = 'F')

            if self.verbosity > 2: print("fitness...")
            F_offspring = calc_fitness(X_offspring,y_t,self.fit_choice)
            # F_offspring = parallel(delayed(f[self.fit_choice])(y_t,yhat) for yhat in X_offspring)
            # print("fitnesses:",fitnesses)
            # Assign fitnesses to inidividuals in population
            for ind, fit in zip(offspring, F_offspring):
                if isinstance(fit,(list,np.ndarray)): # calc_fitness returned raw fitness values
                    fit[fit < 0] = self.max_fit
                    fit[np.isnan(fit)] = self.max_fit
                    fit[np.isinf(fit)] = self.max_fit
                    ind.fitness_vec = fit
                    ind.fitness = np.mean(ind.fitness_vec)
                else:
                    # print("fit.shape:",fit.shape)
                    ind.fitness = np.nanmin([fit,self.max_fit])
            # if self.verbosity > 0: print("median fitness offspring: %0.2f" % np.median([x.fitness for x in offspring]))

            # Survival the next generation individuals
            if self.verbosity > 2: print("survival..")
            if self.sel == 'tournament':
                survivors, survivor_index = tournament(pop.individuals + offspring, self.tourn_size, num_selections = len(pop.individuals))
            elif self.sel == 'lexicase':
                survivors, survivor_index = lexicase(pop.individuals + offspring, num_selections = len(pop.individuals), survival = True)
            elif self.sel == 'epsilon_lexicase':
                survivors, survivor_index = epsilon_lexicase(pop.individuals + offspring, num_selections = len(pop.individuals), survival = True)
            elif self.sel == 'deterministic_crowding':
                survivors, survivor_index = deterministic_crowding(pop.individuals,offspring,pop.X,X_offspring)

            if self.elitism and min([x.fitness for x in survivors]) > elite.fitness:
                # if the elite individual did not survive and elitism is on, replace worst individual with elite
                rep_index = np.argmax([x.fitness for x in survivors])
                survivors[rep_index] = elite
                survivor_index[rep_index] = elite_index
            # print("current population:",stacks_2_eqns(pop.individuals))
            # print("current pop.X:",pop.X[:,:4])
            # print("offspring:",stacks_2_eqns(offspring))
            # print("current X_offspring:",X_offspring[:,:4])
            # print("survivor index:",survivor_index)
            # print("survivors:",stacks_2_eqns(survivors))
            pop.individuals[:] = survivors
            pop.X = np.vstack((pop.X, X_offspring))[survivor_index,:]

            # if pop.X.shape[0] != self.population_size:
            #     pdb.set_trace()
            # print("new pop.X:",pop.X[:,:4])
            # pdb.set_trace()
            # pop.X = pop.X[survivor_index,:]
            #[[s for s in survivor_index if s<len(pop.individuals)],:],
                                    #  X_offspring[[s-len(pop.individuals) for s in survivor_index if s>=len(pop.individuals)],:]))
            if self.verbosity > 2: print("median fitness survivors: %0.2f" % np.median([x.fitness for x in pop.individuals]))

            pbar.set_description('Internal CV: {:1.3f}'.format(self._best_score))
            pbar.update(1)
        # end of main GP loop
            ####################
        if self.verbosity > 0: print('finished. best internal val score: {:1.3f}'.format(self._best_score))
        if self.verbosity > 0: print("final model:\n",self.print_model())

        return self

    def transform(self,x,inds=None,labels = None):
        """return a transformation of x using population outputs"""
        if inds:
            return np.asarray([out(I,x,labels,self.otype) for I in inds],order='F')
            # return np.asarray(list(map(lambda I: out(I,x,labels), inds)),order='F')
        else:
            return np.asarray(list(map(lambda I: out(I,x,labels,self.otype), self._best_inds)),order='F')

    def impute_data(self,x):
        """Imputes data set containing Nan values"""
        imp = Imputer(missing_values='NaN', strategy='mean', axis=0)
        return imp.fit_transform(x)

    def clean(self,x):
        """remove nan and inf rows from x"""
        return x[~np.any(np.isnan(x) | np.isinf(x),axis=1)]

    def clean_with_zeros(self,x):
        """ set nan and inf rows from x to zero"""
        x[~np.any(np.isnan(x) | np.isinf(x),axis=1)] = 0
        return x

    def predict(self, testing_features):
        """predict on a holdout data set."""
        # print("best_inds:",self._best_inds)
        # print("best estimator size:",self._best_estimator.coef_.shape)
        if self.clean:
            testing_features = self.impute_data(testing_features)

        if self._best_inds is None:
            return self._best_estimator.predict(testing_features)
        else:
            X_transform = (np.asarray(list(map(lambda I: out(I,testing_features,otype=self.otype), self._best_inds))))
            return self._best_estimator.predict(X_transform[self.valid_loc(self._best_inds),:].transpose())

    def fit_predict(self, features, labels):
        """Convenience function that fits a pipeline then predicts on the provided features

        Parameters
        ----------
        features: array-like {n_samples, n_features}
            Feature matrix
        labels: array-like {n_samples}
            List of class labels for prediction

        Returns
        ----------
        array-like: {n_samples}
            Predicted labels for the provided features

        """
        self.fit(features, labels)
        return self.predict(features)

    def score(self, testing_features, testing_labels):
        """estimates accuracy on testing set"""
        # print("test features shape:",testing_features.shape)
        # print("testing labels shape:",testing_labels.shape)
        yhat = self.predict(testing_features)
        return self.scoring_function(testing_labels,yhat)

    def export(self, output_file_name):
        """exports engineered features

        Parameters
        ----------
        output_file_name: string
            String containing the path and file name of the desired output file

        Returns
        -------
        None

        """
        if self._best_estimator is None:
            raise ValueError('A model has not been optimized. Please call fit() first.')

        # Have the exported code import all of the necessary modules and functions
        # write model form from coefficients and features
        model = ''
        sym_model = ''
        x = 0
        for c,i in zip(self._best_estimator.coef_,stacks_2_eqns(self._best_inds)):
           if c != 0:
               if model:
                   model+= '+'
                   sym_model += '+'
               model+= str(c) + '*' + str(i)
               sym_model += "k_" + str(x) + '*' + str(i)
               x += 1

        print_text = "exact_model: " + model
        print_text += "\nsymbolic_model: " + sym_model
        print_text += "\ncoefficients: " + str([c for c in self._best_estimator.coef_ if c != 0])
        print_text += "\nfeatures: " + str([s for s,c in zip(stacks_2_eqns(self._best_inds),self._best_estimator.coef_) if c!=0])

        with open(output_file_name, 'w') as output_file:
            output_file.write(print_text)

    def init_pop(self,num_features=1):
        """initializes population of features as GP stacks."""
        pop = Pop(self.population_size,num_features)
        # make programs
        if self.seed_with_ml:
            # initial population is the components of the default ml model
            if type(self.ml) == type(LassoLarsCV()):
                # add all model components with non-zero coefficients
                for i,(c,p) in enumerate(it.zip_longest([c for c in self.ml.coef_ if c !=0],pop.individuals,fillvalue=None)):
                    if c is not None and p is not None:
                        p.stack = [node('x',loc=i)]
                    elif p is not None:
                        # make program if pop is bigger than model componennts
                        make_program(p.stack,self.func_set,self.term_set,np.random.randint(self.min_depth,self.max_depth+1),self.otype)
                        p.stack = list(reversed(p.stack))
            else: # seed with raw features
                # if list(self.ml.coef_):
                #pdb.set_trace()
                try:
                    if self.population_size < self.ml.coef_.shape[0]:
                        # seed pop with highest coefficients
                        coef_order = np.argsort(self.ml.coef_[::-1])
                        for c,p in zip(coef_order,pop.individuals):
                            p.stack = [node('x',loc=c)]
                    else:
                        raise(AttributeError)
                except Exception: # seed pop with raw features
                     for i,p in it.zip_longest(range(self._training_features.shape[1]),pop.individuals,fillvalue=None):
                        if p is not None:
                            if i is not None:
                                p.stack = [node('x',loc=i)]
                            else:
                                make_program(p.stack,self.func_set,self.term_set,np.random.randint(self.min_depth,self.max_depth+1),self.otype)
                                p.stack = list(reversed(p.stack))

            # print initial population
            if self.verbosity > 2: print("seeded initial population:",stacks_2_eqns(pop.individuals))


        else:
            for I in pop.individuals:
                depth = np.random.randint(self.min_depth,self.max_depth+1)
                # print("hex(id(I)):",hex(id(I)))
                # depth = 2;
                # print("initial I.stack:",I.stack)

                make_program(I.stack,self.func_set,self.term_set,depth,self.otype)
                # print(I.stack)
                I.stack = list(reversed(I.stack))

            # print(I.stack)

        return pop

    def print_model(self):
        """prints model contained in best inds, if ml has a coefficient property.
        otherwise, prints the features generated by FEW."""
        model = ''
        if self._best_inds:
            # pdb.set_trace()
            # if hasattr(self.ml,'coef_'):
            try:
                if self.ml.coef_.shape[0]==1 or len(self.ml.coef_.shape)==1:
                    if self.ml.coef_.shape[0]==1:
                        s = np.argsort(np.abs(self.ml.coef_[0]))[::-1]
                        scoef = self.ml.coef_[0][s]
                    else:
                        s = np.argsort(np.abs(self.ml.coef_))[::-1]
                        scoef = self.ml.coef_[s]
                    bi = [self._best_inds[k] for k in s]
                    model = ' +\n'.join([str(round(c,3))+'*'+stack_2_eqn(f) for i,(f,c) in enumerate(zip(bi,scoef)) if round(scoef[i],3) != 0])
                else:
                    # more than one decision function is fit. print all.
                    for j,coef in enumerate(self.ml.coef_):
                        s = np.argsort(np.abs(coef))[::-1]
                        scoef = coef[s]
                        bi =[self._best_inds[k] for k in s]
                        model += '\nclass'+str(j)+' :'+' + '.join([str(round(c,3))+'*'+stack_2_eqn(f) for i,(f,c) in enumerate(zip(bi,coef)) if coef[i] != 0])

            except:
                return stacks_2_eqns(self._best_inds)
        else:
            return 'original features'
        return model

    def model(self):
        """return stacks_2_eqns output"""
        return stacks_2_eqns(self._best_inds)

    def valid_loc(self,individuals):
        """returns the indices of individuals with valid fitness."""

        return [index for index,i in enumerate(individuals) if i.fitness < self.max_fit and i.fitness >= 0]

    def valid(self,individuals):
        """returns the sublist of individuals with valid fitness."""

        return [i for i in individuals if i.fitness < self.max_fit and i.fitness >= 0]

    def get_params(self, deep=None):
        """Get parameters for this estimator

        This function is necessary for FEW to work as a drop-in feature constructor in,
        e.g., sklearn.model_selection.cross_val_score

        Parameters
        ----------
        deep: unused
            Only implemented to maintain interface for sklearn

        Returns
        -------
        params: mapping of string to any
            Parameter names mapped to their values
        """
        return self.params

    def get_diversity(self,X):
        """compute mean diversity of individual outputs"""
        # diversity in terms of cosine distances between features
        feature_correlations = np.zeros(X.shape[0]-1)
        for i in np.arange(1,X.shape[0]-1):
            feature_correlations[i] = max(0.0,r2_score(X[0],X[i]))
        # pdb.set_trace()
        self.diversity.append(1-np.mean(feature_correlations))

def positive_integer(value):
    """Ensures that the provided value is a positive integer; throws an exception otherwise

    Parameters
    ----------
    value: int
        The number to evaluate

    Returns
    -------
    value: int
        Returns a positive integer
    """
    try:
        value = int(value)
    except Exception:
        raise argparse.ArgumentTypeError('Invalid int value: \'{}\''.format(value))
    if value < 0:
        raise argparse.ArgumentTypeError('Invalid positive int value: \'{}\''.format(value))
    return value

def float_range(value):
    """Ensures that the provided value is a float integer in the range (0., 1.); throws an exception otherwise

    Parameters
    ----------
    value: float
        The number to evaluate

    Returns
    -------
    value: float
        Returns a float in the range (0., 1.)
    """
    try:
        value = float(value)
    except:
        raise argparse.ArgumentTypeError('Invalid float value: \'{}\''.format(value))
    if value < 0.0 or value > 1.0:
        raise argparse.ArgumentTypeError('Invalid float value: \'{}\''.format(value))
    return value

# dictionary of ml options
ml_dict = {
        'lasso': LassoLarsCV(),
        'svr': SVR(),
        'lsvr': LinearSVR(),
        'lr': LogisticRegression(solver='sag'),
        'sgd': SGDClassifier(loss='log',penalty='l1'),
        'svc': SVC(),
        'lsvc': LinearSVC(),
        'rfc': RandomForestClassifier(),
        'rfr': RandomForestRegressor(),
        'dtc': DecisionTreeClassifier(),
        'dtr': DecisionTreeRegressor(),
        'dc': DistanceClassifier(),
        'knc': KNeighborsClassifier(),
        'knr': KNeighborsRegressor(),
        None: None
}
# main functions
def main():
    """Main function that is called when FEW is run on the command line"""
    parser = argparse.ArgumentParser(description='A feature engineering wrapper for '
                                                 'machine learning algorithms using genetic programming.',
                                     add_help=False)

    parser.add_argument('INPUT_FILE', type=str, help='Data file to run FEW on; ensure that the target/label column is labeled as "label".')

    parser.add_argument('-h', '--help', action='help', help='Show this help message and exit.')

    parser.add_argument('-is', action='store', dest='INPUT_SEPARATOR', default=None,
                        type=str, help='Character used to separate columns in the input file.')

    parser.add_argument('-o', action='store', dest='OUTPUT_FILE', default='',
                        type=str, help='File to export the final model.')

    parser.add_argument('-g', action='store', dest='GENERATIONS', default=100,
                        type=positive_integer, help='Number of generations to run FEW.')

    parser.add_argument('-p', action='store', dest='POPULATION_SIZE', default=50, #type=positive_integer,
                         help='Number of individuals in the GP population. Follow the number with x to set population size as a multiple of raw feature size.')

    parser.add_argument('-mr', action='store', dest='MUTATION_RATE', default=0.5,
                        type=float_range, help='GP mutation rate in the range [0.0, 1.0].')

    parser.add_argument('-xr', action='store', dest='CROSSOVER_RATE', default=0.5,
                        type=float_range, help='GP crossover rate in the range [0.0, 1.0].')

    parser.add_argument('-ml', action='store', dest='MACHINE_LEARNER', default=None,
                        choices = ['lasso','svr','lsvr','lr','svc','rfc','rfr','dtc','dtr','dc','knn','sgd'],
                        type=str, help='ML algorithm to pair with features. Default: Lasso (regression), LogisticRegression (classification)')

    parser.add_argument('-min_depth', action='store', dest='MIN_DEPTH', default=1,
                        type=positive_integer, help='Minimum length of GP programs.')

    parser.add_argument('-max_depth', action='store', dest='MAX_DEPTH', default=2,
                        type=positive_integer, help='Maximum number of nodes in GP programs.')

    parser.add_argument('-max_depth_init', action='store', dest='MAX_DEPTH_INIT', default=2,
                        type=positive_integer, help='Maximum number of nodes in initialized GP programs.')

    parser.add_argument('-op_weight', action='store', dest='OP_WEIGHT', default=1,
                        type=bool, help='Weight variables for inclusion in synthesized features based on ML scores. Default: off')

    parser.add_argument('-sel', action='store', dest='SEL', default='epsilon_lexicase', choices = ['tournament','lexicase','epsilon_lexicase','deterministic_crowding'],
                        type=str, help='Selection method (Default: tournament)')

    parser.add_argument('-tourn_size', action='store', dest='TOURN_SIZE', default=2,
                        type=positive_integer, help='Tournament size for tournament selection (Default: 2)')

    parser.add_argument('-fit', action='store', dest='FIT_CHOICE', default=None, choices = ['mse','mae','r2','vaf',
                        'mse_rel','mae_rel','r2_rel','vaf_rel'],
                        type=str, help='Fitness metric (Default: dependent on ml used)')

    parser.add_argument('--no_seed', action='store_false', dest='SEED_WITH_ML', default=True,
                    help='Flag to NOT seed initial GP population with components of the ML model.')

    parser.add_argument('--elitism', action='store_true', dest='ELITISM', default=False,
                    help='Flag to force survival of best individual in GP population.')

    parser.add_argument('--erc', action='store_true', dest='ERC', default=False,
                    help='Flag to use ephemeral random constants in GP feature construction.')

    parser.add_argument('--bool', action='store_true', dest='BOOLEAN', default=False,
                    help='Flag to include boolean operators when constructing features.')

    parser.add_argument('-otype', action='store', dest='OTYPE', default='f',choices=['f','b'],
                    type=str,help='Feature output type. f: float, b: boolean.')

    parser.add_argument('--class', action='store_true', dest='CLASSIFICATION', default=False,
                    help='Flag to conduct clasisfication rather than regression.')

    parser.add_argument('--mdr', action='store_true',dest='MDR',default=False,
                    help='Flag to use MDR nodes.')

    parser.add_argument('--diversity', action='store_true', dest='TRACK_DIVERSITY', default=False,
                    help='Flag to store diversity of feature transforms each generation.')

    parser.add_argument('--clean', action='store_true', dest='CLEAN', default=False,
                    help='Flag to clean input data of missing values.')

    parser.add_argument('-s', action='store', dest='RANDOM_STATE', default=np.random.randint(4294967295),
                        type=int, help='Random number generator seed for reproducibility. Note that using multi-threading may '
                                       'make exacts results impossible to reproduce.')

    parser.add_argument('-v', action='store', dest='VERBOSITY', default=1, choices=[0, 1, 2, 3],
                        type=int, help='How much information FEW communicates while it is running: 0 = none, 1 = minimal, 2 = lots, 3 = all.')

    parser.add_argument('--no-update-check', action='store_true', dest='DISABLE_UPDATE_CHECK', default=False,
                        help='Flag indicating whether the FEW version checker should be disabled.')

    parser.add_argument('--version', action='version', version='FEW {version}'.format(version=__version__),
                        help='Show FEW\'s version number and exit.')

    args = parser.parse_args()

    # if args.VERBOSITY >= 2:
    #     print('\nFEW settings:')
    #     for arg in sorted(args.__dict__):
    #         if arg == 'DISABLE_UPDATE_CHECK':
    #             continue
    #         print('{}\t=\t{}'.format(arg, args.__dict__[arg]))
    #     print('')

    # load data from csv file
    if args.INPUT_SEPARATOR is None:
        input_data = pd.read_csv(args.INPUT_FILE, sep=args.INPUT_SEPARATOR,engine='python')
    else: # use c engine for read_csv is separator is specified
        input_data = pd.read_csv(args.INPUT_FILE, sep=args.INPUT_SEPARATOR)

    # if 'Label' in input_data.columns.values:
    input_data.rename(columns={'Label': 'label','Class':'label','class':'label','target':'label'}, inplace=True)

    RANDOM_STATE = args.RANDOM_STATE if args.RANDOM_STATE > 0 else None

    train_i, test_i = train_test_split(input_data.index,
                                                        stratify = None,#  stratify=input_data['label'].values,
                                                         train_size=0.75,
                                                         test_size=0.25,
                                                         random_state=RANDOM_STATE)

    training_features = input_data.loc[train_i].drop('label', axis=1).values
    training_labels = input_data.loc[train_i, 'label'].values

    testing_features = input_data.loc[test_i].drop('label', axis=1).values
    testing_labels = input_data.loc[test_i, 'label'].values

    learner = FEW(generations=args.GENERATIONS, population_size=args.POPULATION_SIZE,
                mutation_rate=args.MUTATION_RATE, crossover_rate=args.CROSSOVER_RATE,
                ml = ml_dict[args.MACHINE_LEARNER], min_depth = args.MIN_DEPTH,
                max_depth = args.MAX_DEPTH, sel = args.SEL, tourn_size = args.TOURN_SIZE,
                seed_with_ml = args.SEED_WITH_ML, op_weight = args.OP_WEIGHT,
                erc = args.ERC, random_state=args.RANDOM_STATE, verbosity=args.VERBOSITY,
                disable_update_check=args.DISABLE_UPDATE_CHECK,fit_choice = args.FIT_CHOICE,
                boolean=args.BOOLEAN,classification=args.CLASSIFICATION,clean = args.CLEAN,
                track_diversity=args.TRACK_DIVERSITY,mdr=args.MDR,otype=args.OTYPE)

    learner.fit(training_features, training_labels)
    # pdb.set_trace()
    if args.VERBOSITY >= 1:
        print('\nTraining accuracy: {:1.3f}'.format(learner.score(training_features, training_labels)))
        print('Test accuracy: {:1.3f}'.format(learner.score(testing_features, testing_labels)))

    if args.OUTPUT_FILE != '':
        learner.export(args.OUTPUT_FILE)


if __name__ == '__main__':
    main()



"""
Copyright 2016 William La Cava

This file is part of the FEW library.

The FEW library is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

The FEW library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
the FEW library. If not, see http://www.gnu.org/licenses/.

"""
# unit tests for evaluation methods.
from few import FEW
from few.evaluation import *
from few.population import *
from sklearn.datasets import load_boston
import numpy as np
def test_out_shapes():
    """test_evaluation.py: program output is correct size """
    # load test data set
    boston = load_boston()
    # boston.data = boston.data[::10]
    # boston.target = boston.target[::10]
    n_features = boston.data.shape[1]
    # function set

    # terminal set
    term_set = []
    # numbers represent column indices of features
    for i in np.arange(n_features):
        term_set.append(node('x',loc=i)) # features
        # term_set.append(('k',0,np.random.rand())) # ephemeral random constants

    # initialize population
    pop_size = 5;
    few = FEW(population_size=pop_size,seed_with_ml=False)
    few.term_set = term_set
    pop = few.init_pop(n_features)

    pop.X = np.asarray(list(map(lambda I: out(I,boston.data), pop.individuals)))

    #pop.X = out(pop.individuals[0],boston.data,boston.target)
    print("pop.X.shape:",pop.X.shape)
    print("boston.target.shape",boston.target.shape)
    assert pop.X.shape == (pop_size, boston.target.shape[0])

def test_out_is_correct():
    """test_evaluation.py: output matches known function outputs """

    boston = load_boston()
    n_features = boston.data.shape[1]
    X = boston.data
    Y = boston.target
    p1 = Ind()
    p2 = Ind()
    p3 = Ind()
    p4 = Ind()
    p5 = Ind()

    p1.stack = [node('x',loc=4), node('x',loc=5), node('-'),
                node('k',value=0.175), node('log'), node('-')]

    p2.stack = [node('x',loc=7),node('x',loc=8),node('*')]

    p3.stack =  [node('x',loc=0),node('exp'), node('x',loc=5),node('x',loc=7),
                node('*'),node('/')]

    p4.stack =  [node('x',loc=12),node('sin')]

    p5.stack = [node('k',value=178.3),node('x',loc=8),node('*'),node('x',loc=7),node('cos'),node('+')]

    y1 = safe(np.log(0.175) - (X[:,5] - X[:,4]))
    y2 = safe(X[:,7]*X[:,8])
    y3 = safe(divs(X[:,5]*X[:,7],np.exp(X[:,0])))
    y4 = safe(np.sin(X[:,12]))
    y5 = safe(178.3*X[:,8]+np.cos(X[:,7]))

    # y1,y2,y3,y4,y5 = safe(y1),safe(y2),safe(y3),safe(y4),safe(y5)

    assert np.array_equal(y1,out(p1,X))
    print("y1 passed")
    assert np.array_equal(y2,out(p2,X))
    print("y2 passed")
    assert np.array_equal(y3, out(p3,X))
    print("y3 passed")
    # print("y4:",y4,"y4hat:",out(p4,X,Y))
    assert np.array_equal(y4, out(p4,X))
    print("y4 passed")
    assert np.array_equal(y5, out(p5,X))

def test_calc_fitness_shape():
    """test_evaluation.py: calc_fitness correct shapes """
    # load test data set
    boston = load_boston()
    # boston.data = boston.data[::10]
    # boston.target = boston.target[::10]
    n_features = boston.data.shape[1]
    # terminal set
    term_set = []
    # numbers represent column indices of features
    for i in np.arange(n_features):
        term_set.append(node('x',loc=i)) # features
        # term_set.append(('k',0,np.random.rand())) # ephemeral random constants

    # initialize population
    pop_size = 5;
    few = FEW(population_size=pop_size,seed_with_ml=False)
    few.term_set = term_set
    pop = few.init_pop(n_features)

    pop.X = np.asarray(list(map(lambda I: out(I,boston.data), pop.individuals)))

    fitnesses = calc_fitness(pop.X,boston.target,'mse')
    assert len(fitnesses) == len(pop.individuals)

    # test vectorized fitnesses
    vec_fitnesses = calc_fitness(pop.X,boston.target,'mse_vec')
    fitmat = np.asarray(vec_fitnesses)
    print("fitmat.shape:",fitmat.shape)
    assert fitmat.shape == (len(pop.individuals),boston.target.shape[0])



"""
Copyright 2016 William La Cava

This file is part of the FEW library.

The FEW library is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

The FEW library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
the FEW library. If not, see http://www.gnu.org/licenses/.

"""
import numpy as np
from few import FEW
from few.population import *
from itertools import accumulate
# unit tests for population methods.
def test_pop_shape():
    """test_population.py: population class returns correct sizes """
    # pop = Pop(0)
    # assert len(pop) == 0
    pop = Pop(10)
    assert len(pop.individuals) == 10
    print("pop.X.shape",pop.X.shape)
    assert pop.X.shape == (1,10)
    assert pop.E.shape == (0,)


    pop = Pop(73)
    assert len(pop.individuals) == 73
    assert pop.X.shape == (1,73)
    assert pop.E.shape == (0,)

    pop = Pop(73,5)
    assert len(pop.individuals) == 73
    assert pop.X.shape == (5,73)
    assert pop.E.shape == (0,)

# NOTE: popultation initialization is done in the ellyn class now. this test needs to be rewritten.
def test_pop_init():
    """test_population.py: population initialization makes valid trees """
    # define function set
    # function set
    # func_set = [('+',2),('-',2),('*',2),('/',2),('sin',1),('cos',1),('exp',1),('log',1)]
    # terminal set
    term_set = []
    n_features = 3
    # numbers represent column indices of features
    term_set = [node('x',loc=i) for i in np.arange(n_features)]
        # term_set.append(('erc',0,np.random.rand())) # ephemeral random constants
    few = FEW(seed_with_ml=False)
    few.term_set = term_set

    pop = few.init_pop(n_features)

    for I in pop.individuals:
        assert is_valid_program(I.stack)

def is_valid_program(p):
    """ checks that the accumulated program length is always greater than the
    accumulated arities, indicating that the appropriate number of arguments is
    alway present for functions. It then checks that the sum of arties +1
    exactly equals the length of the stack, indicating that there are no
    missing arguments. """
    # print("p:",p)
    arities = list(a.arity[a.in_type] for a in p)
    accu_arities = list(accumulate(arities))
    accu_len = list(np.arange(len(p))+1)
    check = list(a < b for a,b in zip(accu_arities,accu_len))
    # print("accu_arities:",accu_arities)
    # print("accu_len:",accu_len)
    # print("accu_arities < accu_len:",accu_arities<accu_len)
    return all(check) and sum([a.arity[a.in_type] for a in p]) +1 == len(p) and len(p)>0



# -*- coding: utf-8 -*-
"""
Copyright 2016 William La Cava

license: GNU/GPLv3

"""
import numpy as np
from .population import make_program
from itertools import accumulate
import pdb

# from few.tests.test_population import is_valid_program

def cross(p_i,p_j, max_depth = 2):
    """subtree-like swap crossover between programs p_i and p_j."""
    # only choose crossover points for out_types available in both programs
    # pdb.set_trace()
    # determine possible outttypes
    types_p_i = [t for t in [p.out_type for p in p_i]]
    types_p_j = [t for t in [p.out_type for p in p_j]]
    types = set(types_p_i).intersection(types_p_j)

    # grab subtree of p_i
    p_i_sub = [i for i,n in enumerate(p_i) if n.out_type in types]
    x_i_end = np.random.choice(p_i_sub)
    x_i_begin = x_i_end
    arity_sum = p_i[x_i_end].arity[p_i[x_i_end].in_type]
    # print("x_i_end:",x_i_end)
    # i = 0
    while (arity_sum > 0): #and i < 1000:
        if x_i_begin == 0:
            print("arity_sum:",arity_sum,"x_i_begin:",x_i_begin,"x_i_end:",x_i_end)
        x_i_begin -= 1
        arity_sum += p_i[x_i_begin].arity[p_i[x_i_begin].in_type]-1
    #     i += 1
    # if i == 1000:
    #     print("in variation")
    #     pdb.set_trace()

    # grab subtree of p_j with matching out_type to p_i[x_i_end]
    p_j_sub = [i for i,n in enumerate(p_j) if n.out_type == p_i[x_i_end].out_type]
    x_j_end = np.random.choice(p_j_sub)
    x_j_begin = x_j_end
    arity_sum = p_j[x_j_end].arity[p_j[x_j_end].in_type]
    # i = 0
    while (arity_sum > 0): #and i < 1000:
        if x_j_begin == 0:
            print("arity_sum:",arity_sum,"x_j_begin:",x_j_begin,"x_j_end:",x_j_end)
            print("p_j:",p_j)
        x_j_begin -= 1
        arity_sum += p_j[x_j_begin].arity[p_j[x_j_begin].in_type]-1
    #     i += 1
    # if i == 1000:
    #     print("in variation")
    #     pdb.set_trace()
    #swap subtrees
    tmpi = p_i[:]
    tmpj = p_j[:]
    tmpi[x_i_begin:x_i_end+1:],tmpj[x_j_begin:x_j_end+1:] = tmpj[x_j_begin:x_j_end+1:],tmpi[x_i_begin:x_i_end+1:]

    if not is_valid_program(p_i) or not is_valid_program(p_j):
        # pdb.set_trace()
        print("parent 1:",p_i,"x_i_begin:",x_i_begin,"x_i_end:",x_i_end)
        print("parent 2:",p_j,"x_j_begin:",x_j_begin,"x_j_end:",x_j_end)
        print("child 1:",tmpi)
        print("child 2:",tmpj)
        raise ValueError('Crossover produced an invalid program.')

    # size check, then assignment
    if len(tmpi) <= 2**max_depth-1:
        p_i[:] = tmpi
    if len(tmpj) <= 2**max_depth-1:
        p_j[:] = tmpj


def mutate(p_i,func_set,term_set): #, max_depth=2
    """point mutation, addition, removal"""
    point_mutate(p_i,func_set,term_set)

def point_mutate(p_i,func_set,term_set):
    """point mutation on individual p_i"""
    # point mutation
    x = np.random.randint(len(p_i))
    arity = p_i[x].arity[p_i[x].in_type]
    # find eligible replacements based on arity and type
    reps = [n for n in func_set+term_set
            if n.arity[n.in_type]==arity and n.out_type==p_i[x].out_type
            and n.in_type==p_i[x].in_type]

    tmp = reps[np.random.randint(len(reps))]
    tmp_p = p_i[:]
    p_i[x] = tmp
    if not is_valid_program(p_i):
        print("old:",tmp_p)
        print("new:",p_i)
        raise ValueError('Mutation produced an invalid program.')
#
# def add_mutate(p_i,func_set,term_set, max_depth=2):
#     """ mutation that adds operation to program"""
#     #choose node. move it down, pick an operator to put before it, with another leaf if
#     #new operator arity requires it.
#     #make sure size requirements are not invalidated (if they are, discard changes)
#
# def sub_mutate(p_i,func_set,term_set, max_depth=2):
#     """ mutation that removes operation from program"""
#     #choose a node with arity>0. replace it and its subtree with a node with lower arity.

def is_valid_program(p):
    """checks whether program p makes a syntactically valid tree.

    checks that the accumulated program length is always greater than the
    accumulated arities, indicating that the appropriate number of arguments is
    alway present for functions. It then checks that the sum of arties +1
    exactly equals the length of the stack, indicating that there are no
    missing arguments.
    """
    # print("p:",p)
    arities = list(a.arity[a.in_type] for a in p)
    accu_arities = list(accumulate(arities))
    accu_len = list(np.arange(len(p))+1)
    check = list(a < b for a,b in zip(accu_arities,accu_len))
    # print("accu_arities:",accu_arities)
    # print("accu_len:",accu_len)
    # print("accu_arities < accu_len:",accu_arities<accu_len)
    return all(check) and sum(a.arity[a.in_type] for a in p) +1 == len(p) and len(p)>0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

def calculate_version():
    initpy = open('few/_version.py').read().split('\n')
    version = list(filter(lambda x: '__version__' in x, initpy))[0].split('\'')[1]
    return version

package_version = calculate_version()

setup(
    name='FEW',
    version=package_version,
    author='William La Cava',
    author_email='lacava@upenn.edu',
    packages=find_packages(),
    url='https://github.com/lacava/few',
    download_url='https://github.com/lacava/few/releases/tag/'+package_version,
    license='GNU/GPLv3',
    entry_points={'console_scripts': ['few=few:main', ]},
    test_suite='nose.collector',
    tests_require=['nose'],
    description=('Feature Engineering Wrapper'),
    long_description='''
A feature engineering wrapper for scikitlearn based on genetic programming.

Contact:
===
e-mail: lacava@upenn.edu

This project is hosted at https://github.com/lacava/few
''',
    zip_safe=True,
    install_requires=['numpy', 'scipy', 'pandas', 'scikit-learn', 'update_checker', 'tqdm', 'joblib','DistanceClassifier', 'scikit-mdr'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        # 'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    keywords=['data science', 'machine learning', 'classification'],
)



# -*- coding: utf-8 -*-
"""
Copyright 2016 William La Cava

license: GNU/GPLv3

"""
import numpy as np
from sklearn.metrics import explained_variance_score, mean_absolute_error, mean_squared_error, median_absolute_error, r2_score
import pdb
from sklearn.metrics import silhouette_samples, silhouette_score, accuracy_score


# evaluation functions. these can be sped up using a GPU!

eval_dict = {
# float operations
    '+': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() + stack_float.pop(),
    '-': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() - stack_float.pop(),
    '*': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() * stack_float.pop(),
    '/': lambda n,features,stack_float,stack_bool,labels: divs(stack_float.pop(),stack_float.pop()),
    'sin': lambda n,features,stack_float,stack_bool,labels: np.sin(stack_float.pop()),
    'cos': lambda n,features,stack_float,stack_bool,labels: np.cos(stack_float.pop()),
    'exp': lambda n,features,stack_float,stack_bool,labels: np.exp(stack_float.pop()),
    'log': lambda n,features,stack_float,stack_bool,labels: logs(stack_float.pop()),#np.log(np.abs(stack_float.pop())),
    'x':  lambda n,features,stack_float,stack_bool,labels: features[:,n.loc],
    'k': lambda n,features,stack_float,stack_bool,labels: np.ones(features.shape[0])*n.value,
    '^2': lambda n,features,stack_float,stack_bool,labels: stack_float.pop()**2,
    '^3': lambda n,features,stack_float,stack_bool,labels: stack_float.pop()**3,
    'sqrt': lambda n,features,stack_float,stack_bool,labels: np.sqrt(np.abs(stack_float.pop())),
    # 'rbf': lambda n,features,stack_float,stack_bool,labels: np.exp(-(np.norm(stack_float.pop()-stack_float.pop())**2)/2)
# bool operations
    '!': lambda n,features,stack_float,stack_bool,labels: np.logical_not(stack_bool.pop()),
    '&': lambda n,features,stack_float,stack_bool,labels: np.logical_and(stack_bool.pop(), stack_bool.pop()),
    '|': lambda n,features,stack_float,stack_bool,labels: np.logical_or(stack_bool.pop(), stack_bool.pop()),
    '==': lambda n,features,stack_float,stack_bool,labels: stack_bool.pop() == stack_bool.pop(),
    '>_f': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() > stack_float.pop(),
    '<_f': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() < stack_float.pop(),
    '>=_f': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() >= stack_float.pop(),
    '<=_f': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() <= stack_float.pop(),
    '>_b': lambda n,features,stack_float,stack_bool,labels: stack_bool.pop() > stack_bool.pop(),
    '<_b': lambda n,features,stack_float,stack_bool,labels: stack_bool.pop() < stack_bool.pop(),
    '>=_b': lambda n,features,stack_float,stack_bool,labels: stack_bool.pop() >= stack_bool.pop(),
    '<=_b': lambda n,features,stack_float,stack_bool,labels: stack_bool.pop() <= stack_bool.pop(),
    'xor_b': lambda n,features,stack_float,stack_bool,labels: np.logical_xor(stack_bool.pop(),stack_bool.pop()),
    'xor_f': lambda n,features,stack_float,stack_bool,labels: np.logical_xor(stack_float.pop().astype(bool), stack_float.pop().astype(bool)),
# MDR
    'mdr2': lambda n,features,stack_float,stack_bool,labels: n.evaluate(n,stack_float,labels),
# control flow:
    # 'if': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() if stack_bool.pop(),
    # 'ife': lambda n,features,stack_float,stack_bool,labels: stack_float.pop() if stack_bool.pop() else stack_float.pop(),
    }

f = { # available fitness metrics
'mse': lambda y,yhat: mean_squared_error(y,yhat),
'mae': lambda y,yhat: mean_absolute_error(y,yhat),
'mdae': lambda y,yhat: median_absolute_error(y,yhat),
'r2':  lambda y,yhat: 1-r2_score(y,yhat),
'vaf': lambda y,yhat: 1-explained_variance_score(y,yhat),
'silhouette': lambda y,yhat: 1 - silhouette_score(yhat.reshape(-1,1),y),
'accuracy': lambda y,yhat: 1 - accuracy_score(yhat,y),
# non-aggregated fitness calculations
'mse_vec': lambda y,yhat: (y - yhat) ** 2, #mean_squared_error(y,yhat,multioutput = 'raw_values'),
'mae_vec': lambda y,yhat: np.abs(y-yhat), #mean_absolute_error(y,yhat,multioutput = 'raw_values'),
# 'mdae_vec': lambda y,yhat: median_absolute_error(y,yhat,multioutput = 'raw_values'),
'r2_vec':  lambda y,yhat: 1-r2_score_vec(y,yhat),
'vaf_vec': lambda y,yhat: 1-explained_variance_score(y,yhat,multioutput = 'raw_values'),
'silhouette_vec': lambda y,yhat: 1 - silhouette_samples(yhat.reshape(-1,1),y),
'accuracy_vec': lambda y,yhat: 1 - np.sum(yhat==y)/y.shape[0]
}

def safe(x):
    """removes nans and infs from outputs."""
    x[np.isinf(x)] = 1
    x[np.isnan(x)] = 1
    return x

def divs(x,y):
    """safe division"""
    tmp = np.ones(x.shape)
    nonzero_y = y != 0
    tmp[nonzero_y] = x[nonzero_y]/y[nonzero_y]
    return tmp

def logs(x):
    """safe log"""
    tmp = np.ones(x.shape)
    nonzero_x = x != 0
    tmp[nonzero_x] = np.log(np.abs(x[nonzero_x]))
    return tmp


def evaluate(n, features, stack_float, stack_bool,labels=None):
    """evaluate node in program"""
    np.seterr(all='ignore')
    if len(stack_float) >= n.arity['f'] and len(stack_bool) >= n.arity['b']:
        if n.out_type == 'f':
            stack_float.append(safe(eval_dict[n.name](n,features,stack_float,stack_bool,labels)))
            if np.isnan(stack_float[-1]).any() or np.isinf(stack_float[-1]).any():
                print("problem operator:",n)
        else:
            stack_bool.append(safe(eval_dict[n.name](n,features,stack_float,stack_bool,labels)))
            if np.isnan(stack_bool[-1]).any() or np.isinf(stack_bool[-1]).any():
                print("problem operator:",n)

def all_finite(X):
    """returns true if X is finite, false, otherwise"""
    # Adapted from sklearn utils: _assert_all_finite(X)
    # First try an O(n) time, O(1) space solution for the common case that
    # everything is finite; fall back to O(n) space np.isfinite to prevent
    # false positives from overflow in sum method.
    if (X.dtype.char in np.typecodes['AllFloat'] and not np.isfinite(X.sum())
        and not np.isfinite(X).all()):
        return False
    return True

def out(I,features,labels=None,otype='f'):
    """computes the output for individual I"""
    stack_float = []
    stack_bool = []
    # print("stack:",I.stack)
    # evaulate stack over rows of features,labels
    # pdb.set_trace()
    for n in I.stack:
        evaluate(n,features,stack_float,stack_bool,labels)
        # print("stack_float:",stack_float)
    if otype=='f':
        return stack_float[-1] if all_finite(stack_float[-1]) else np.zeros(len(features))
    else:
        return stack_bool[-1] if all_finite(stack_bool[-1]) else np.zeros(len(features))

def calc_fitness(X,labels,fit_choice):
    """computes fitness of individual output yhat.
    yhat: output of a program.
    labels: correct outputs
    fit_choice: choice of fitness function
    """



    return list(map(lambda yhat: f[fit_choice](labels,yhat),X))

def r2_score_vec(y_true,y_pred):
    """ returns non-aggregate version of r2 score.

    based on r2_score() function from sklearn (http://sklearn.org)
    """

    numerator = (y_true - y_pred) ** 2
    denominator = (y_true - np.average(y_true)) ** 2

    nonzero_denominator = denominator != 0
    nonzero_numerator = numerator != 0
    valid_score = nonzero_denominator & nonzero_numerator
    output_scores = np.ones([y_true.shape[0]])
    output_scores[valid_score] = 1 - (numerator[valid_score] /
                                      denominator[valid_score])
    # arbitrary set to zero to avoid -inf scores, having a constant
    # y_true is not interesting for scoring a regression anyway
    output_scores[nonzero_numerator & ~nonzero_denominator] = 0.

    return output_scores



# -*- coding: utf-8 -*-
"""
Copyright 2016 William La Cava

license: GNU/GPLv3

"""
import numpy as np
import copy
import pdb
import uuid
from mdr import MDR
from collections import defaultdict

eqn_dict = {
    '+': lambda n,stack_eqn: '(' + stack_eqn.pop() + '+' + stack_eqn.pop() + ')',
    '-': lambda n,stack_eqn: '(' + stack_eqn.pop() + '-' + stack_eqn.pop()+ ')',
    '*': lambda n,stack_eqn: '(' + stack_eqn.pop() + '*' + stack_eqn.pop()+ ')',
    '/': lambda n,stack_eqn: '(' + stack_eqn.pop() + '/' + stack_eqn.pop()+ ')',
    'sin': lambda n,stack_eqn: 'sin(' + stack_eqn.pop() + ')',
    'cos': lambda n,stack_eqn: 'cos(' + stack_eqn.pop() + ')',
    'exp': lambda n,stack_eqn: 'exp(' + stack_eqn.pop() + ')',
    'log': lambda n,stack_eqn: 'log(' + stack_eqn.pop() + ')',
    '^2': lambda n,stack_eqn: '(' + stack_eqn.pop() + '^2)',
    '^3': lambda n,stack_eqn: '(' + stack_eqn.pop() + '^3)',
    'sqrt': lambda n,stack_eqn: 'sqrt(|' + stack_eqn.pop() + '|)',
    # 'rbf': lambda n,stack_eqn: 'exp(-||' + stack_eqn.pop()-stack_eqn.pop() '||^2/2)',
    'x':  lambda n,stack_eqn: 'x_' + str(n.loc),
    'k': lambda n,stack_eqn: str(n.value),
    'mdr2': lambda n,stack_eqn: 'MDR(' + stack_eqn.pop() + ',' + stack_eqn.pop() + ')',
# bool operations
    '!': lambda n,stack_eqn: '(!' + stack_eqn.pop() + ')',
    '&': lambda n,stack_eqn: '(' + stack_eqn.pop() + '&' + stack_eqn.pop() + ')',
    '|': lambda n,stack_eqn: '(' + stack_eqn.pop() + '|' + stack_eqn.pop() + ')',
    '==': lambda n,stack_eqn: '(' + stack_eqn.pop() + '==' + stack_eqn.pop() + ')',
    '>_f': lambda n,stack_eqn: '(' + stack_eqn.pop() + '>' + stack_eqn.pop() + ')',
    '<_f': lambda n,stack_eqn: '(' + stack_eqn.pop() + '<' + stack_eqn.pop() + ')',
    '>=_f': lambda n,stack_eqn: '(' + stack_eqn.pop() + '>=' + stack_eqn.pop() + ')',
    '<=_f': lambda n,stack_eqn: '(' + stack_eqn.pop() + '<=' + stack_eqn.pop() + ')',
    '>_b': lambda n,stack_eqn: '(' + stack_eqn.pop() + '>' + stack_eqn.pop() + ')',
    '<_b': lambda n,stack_eqn: '(' + stack_eqn.pop() + '<' + stack_eqn.pop() + ')',
    '>=_b': lambda n,stack_eqn: '(' + stack_eqn.pop() + '>=' + stack_eqn.pop() + ')',
    '<=_b': lambda n,stack_eqn: '(' + stack_eqn.pop() + '<=' + stack_eqn.pop() + ')',
    'xor_b': lambda n,stack_eqn: '(' + stack_eqn.pop() + ' XOR ' + stack_eqn.pop() + ')',
    'xor_f': lambda n,stack_eqn: '(' + stack_eqn.pop() + ' XOR ' + stack_eqn.pop() + ')',

}
def run_MDR(n,stack_float,labels=None):
    """run utility function for MDR nodes."""
    # need to check that tmp is categorical

    x1 = stack_float.pop()
    x2 = stack_float.pop()
    # check data is categorical
    if len(np.unique(x1))<=3 and len(np.unique(x2))<=3:
        tmp = np.vstack((x1,x2)).transpose()

        if labels is None: # prediction
            return n.model.transform(tmp)[:,0]
        else: # training
            out =  n.model.fit_transform(tmp,labels)[:,0]

            return out
    else:
        return np.zeros(x1.shape[0])

class node(object):
    """node in program"""
    def __init__(self,name,loc=None,value=None):
        """defines properties of a node given its name"""
        self.name = name
        self.arity = {None:0}
        self.arity['f'] = defaultdict(lambda: 0, {
                           'sin':1,'cos':1,'exp':1,'log':1,'^2':1,'^3':1,
                           'sqrt':1,'if':1,
                           'ife':2,'+':2,'-':2,'*':2,'/':2,'>_f':2,'<_f':2,
                           '>=_f':2,'<=_f':2,'xor_f':2,'mdr2':2})[name]

        self.arity['b'] = defaultdict(lambda: 0, {
                            '!':1,'if':1,'ife':1,
                            '&':2,'|':2,'==':2,'>_b':2,'<_b':2,'>=_b':2,
                            '<=_b':2,'xor_b':2})[name]
        self.in_type = {
        # float operations
            '+':'f', '-':'f', '*':'f', '/':'f', 'sin':'f', 'cos':'f', 'exp': 'f',
            'log':'f', 'x':None, 'k':None, '^2':'f', '^3':'f', 'sqrt': 'f',
            # 'rbf': ,
        # bool operations
            '!':'b', '&':'b', '|':'b', '==':'b', '>_f':'f', '<_f':'f', '>=_f':'f',
            '<=_f':'f', '>_b':'b', '<_b':'b', '>=_b':'b', '<=_b':'b','xor_b':'b',
            'xor_f':'f',
        # mixed
            'mdr2':'f','if':('f','b'),'ife':('f','b')
        }[name]
        self.out_type = {
        # float operations
            '+': 'f','-': 'f','*': 'f','/': 'f','sin': 'f','cos': 'f','exp': 'f',
            'log': 'f','x':'f','k': 'f','^2': 'f','^3': 'f','sqrt': 'f',
            # 'rbf': ,
        # bool operations
            '!': 'b', '&': 'b','|': 'b','==': 'b','>_f': 'b','<_f': 'b','>=_f': 'b',
            '<=_f': 'b','>_b': 'b','<_b': 'b','>=_b': 'b','<=_b': 'b','xor_f':'b',
            'xor_b':'b',
        # mixed
            'mdr2':'b','if':'f','ife':'f'
        }[name]

        if 'mdr' in self.name:
            self.model = MDR()
            self.evaluate = run_MDR

        self.loc = loc
        self.value = value


class Ind(object):
    """class for features, represented as GP stacks."""

    def __init__(self,fitness = -1.0,stack = None):
        """initializes empty individual with invalid fitness."""
        self.fitness = fitness
        self.fitness_vec = []
        self.fitness_bool = []
        self.parentid = []
        self.id = uuid.uuid4()
        if stack is None:
            self.stack = []
        else:
            self.stack = copy.deepcopy(stack)

class Pop(object):
    """class representing population"""
    def __init__(self,pop_size=100,n_samples=1, fit = None):
        """initializes population of inds of size pop_size"""

        self.individuals = []
        # initialize empty output matrix
        self.X = np.empty([n_samples,pop_size],dtype=float,order='F')
        # initialize empty error matrix
        self.E = np.empty(0) #np.empty([n_samples,pop_size],dtype=float)
        # initialize population programs
        for i in np.arange(pop_size):
            if fit is None:
                self.individuals.append(Ind())
            else:
                self.individuals.append(Ind(fitness = fit))

def stacks_2_eqns(stacks):
    """returns equation strings from stacks"""
    if stacks:
        return list(map(lambda p: stack_2_eqn(p), stacks))
    else:
        return []

def stack_2_eqn(p):
    """returns equation string for program stack"""
    stack_eqn = []
    if p: # if stack is not empty
        for n in p.stack:
            eval_eqn(n,stack_eqn)
        return stack_eqn[-1]
    return []

def eval_eqn(n,stack_eqn):
    if len(stack_eqn) >= n.arity['f']+n.arity['b']:
        stack_eqn.append(eqn_dict[n.name](n,stack_eqn))
        # if any(np.isnan(stack_eqn[-1])) or any(np.isinf(stack_eqn[-1])):
        #     print("problem operator:",n)

def make_program(stack,func_set,term_set,max_d,ntype):
    """makes a program stack"""
    # print("stack:",stack,"max d:",max_d)
    if max_d == 0: #or np.random.rand() < float(len(term_set))/(len(term_set)+len(func_set)):
        ts = [t for t in term_set if t.out_type==ntype]

        if not ts:
            raise ValueError('no ts. ntype:'+ntype+'. term_set out_types:'+','.join([t.out_type for t in term_set]))

        stack.append(ts[np.random.choice(len(ts))])
    else:
        fs = [f for f in func_set if (f.out_type==ntype and (f.in_type=='f' or max_d>1))]
        if len(fs)==0:
            print('ntype:',ntype,'\nfunc_set:',[f.name for f in func_set])
        stack.append(fs[np.random.choice(len(fs))])
        tmp = copy.copy(stack[-1])

        for i in np.arange(tmp.arity['f']):
            make_program(stack,func_set,term_set,max_d-1,'f')
        for i in np.arange(tmp.arity['b']):
            make_program(stack,func_set,term_set,max_d-1,'b')



"""
Copyright 2016 William La Cava

This file is part of the FEW library.

The FEW library is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your option)
any later version.

The FEW library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
the FEW library. If not, see http://www.gnu.org/licenses/.

"""
# unit tests for variation methods.
import numpy as np

from few.tests.test_population import is_valid_program,node
from few.variation import cross, mutate

def test_cross_makes_valid_program():
    """test_variation.py: crossover makes valid programs """
    # np.random.seed(65)
    # I = (a+b)*x
    p1 = [node('x',loc=1),node('x',loc=2),node('+'),node('x',loc=3),node('*')]
    # p1 = [{'name':'x','arity':0,'loc':1,'in_type':None,'out_type':'f'},
    #       {'name':'x','arity':0,'loc':2,'in_type':None,'out_type':'f'},
    #       {'name':'+','arity':2,'in_type':'f','out_type':'f'},
    #       {'name':'x','arity':0,'loc':3,'in_type':None,'out_type':'f'},
    #       {'name':'*','arity':2,'in_type':'f','out_type':'f'}]
    # J = (x/z)-(n*b)
    p2 = [node('x',loc=1), node('x',loc=2), node('/'), node('k',value=3.7), node('x',loc=4), node('*'), node('-')]
    # p2 = [{'name':'x','arity':0,'loc':1,'in_type':None,'out_type':'f'},
    #       {'name':'x','arity':0,'loc':2,'in_type':None,'out_type':'f'},
    #       {'name':'/','arity':2,'in_type':'f','out_type':'f'},
    #       {'name':'k','arity':0,'value':3.7,'in_type':None,'out_type':'f'},
    #       {'name':'x','arity':0,'loc':4,'in_type':None,'out_type':'f'},
    #       {'name':'*','arity':2,'in_type':'f','out_type':'f'},
    #       {'name':'-','arity':2,'in_type':'f','out_type':'f'}]

    for i in np.arange(1000):
        cross(p1,p2)
        assert is_valid_program(p1) and is_valid_program(p2)

def test_mutate_makes_valid_program():
    """test_variation.py: mutation makes valid programs """
    func_set = [node('+'), node('-'), node('*'), node('/'), node('sin'),
                 node('cos'), node('exp'),node('log'), node('^2'),
                 node('^3'), node('sqrt')]
    # terminal set
    term_set = []
    # numbers represent column indices of features
    term_set = [node('x',loc=i) for i in np.arange(10)]
    term_set += [node('k',value=np.random.rand()) for i in np.arange(10)]
    # for i in np.arange(10):
    #     term_set.append({'name':'x','arity':0,'loc':i,'in_type':None,'out_type':'f'}) # features
    #     term_set.append({'name':'k','arity':0,'value':np.random.rand(),'in_type':None,'out_type':'f'}) # ephemeral random constants
    p = [node('k',value=5),node('x',loc=6),node('/'),node('k',value=7),node('x',loc=8),node('*'),node('-')]
    # p = [{'name':'k','arity':0,'value':5,'in_type':None,'out_type':'f'},{'name':'x','arity':0,'loc':6,'in_type':None,'out_type':'f'},{'name':'/','arity':2,'in_type':'f','out_type':'f'},
    #      {'name':'k','arity':0,'value':7,'in_type':None,'out_type':'f'},{'name':'x','arity':0,'loc':8,'in_type':None,'out_type':'f'},
    #      {'name':'*','arity':2,'in_type':'f','out_type':'f'},{'name':'-','arity':2,'in_type':'f','out_type':'f'}]
    for i in np.arange(1000):
        mutate(p,func_set,term_set)
        assert is_valid_program(p)
